@@ -30,31 +30,36 @@ else:
 
 
 class SyntaxHighlighter:
-    _default_style = {
-        "comment": "\x1b[30m\x1b[1m{}\x1b[0m",
-        "keyword": "\x1b[35m\x1b[1m{}\x1b[0m",
-        "builtin": "\x1b[1m{}\x1b[0m",
-        "string": "\x1b[36m{}\x1b[0m",
-        "number": "\x1b[34m\x1b[1m{}\x1b[0m",
-        "operator": "\x1b[35m\x1b[1m{}\x1b[0m",
-        "punctuation": "\x1b[1m{}\x1b[0m",
-        "constant": "\x1b[36m\x1b[1m{}\x1b[0m",
-        "identifier": "\x1b[1m{}\x1b[0m",
-        "other": "{}",
-    }
+    _default_style = frozenset(
+        {
+            "comment": "\x1b[30m\x1b[1m{}\x1b[0m",
+            "keyword": "\x1b[35m\x1b[1m{}\x1b[0m",
+            "builtin": "\x1b[1m{}\x1b[0m",
+            "string": "\x1b[36m{}\x1b[0m",
+            "number": "\x1b[34m\x1b[1m{}\x1b[0m",
+            "operator": "\x1b[35m\x1b[1m{}\x1b[0m",
+            "punctuation": "\x1b[1m{}\x1b[0m",
+            "constant": "\x1b[36m\x1b[1m{}\x1b[0m",
+            "identifier": "\x1b[1m{}\x1b[0m",
+            "other": "{}",
+        }.items()
+    )
 
-    _builtins = set(dir(builtins))
-    _constants = {"True", "False", "None"}
-    _punctation = {"(", ")", "[", "]", "{", "}", ":", ",", ";"}
-    _strings = {tokenize.STRING}
-    _fstring_middle = None
+    _builtins = frozenset(dir(builtins))
+    _constants = frozenset({"True", "False", "None"})
+    _punctuation = frozenset({"(", ")", "[", "]", "{", "}", ":", ",", ";"})
 
     if sys.version_info >= (3, 12):
-        _strings.update({tokenize.FSTRING_START, tokenize.FSTRING_MIDDLE, tokenize.FSTRING_END})
+        _strings = frozenset(
+            {tokenize.STRING, tokenize.FSTRING_START, tokenize.FSTRING_MIDDLE, tokenize.FSTRING_END}
+        )
         _fstring_middle = tokenize.FSTRING_MIDDLE
+    else:
+        _strings = frozenset({tokenize.STRING})
+        _fstring_middle = None
 
     def __init__(self, style=None):
-        self._style = style or self._default_style
+        self._style = style or dict(self._default_style)
 
     def highlight(self, source):
         style = self._style
@@ -79,7 +84,7 @@ class SyntaxHighlighter:
                 else:
                     color = style["identifier"]
             elif type_ == tokenize.OP:
-                if string in self._punctation:
+                if string in self._punctuation:
                     color = style["punctuation"]
                 else:
                     color = style["operator"]
@@ -119,19 +124,21 @@ class SyntaxHighlighter:
 
 
 class ExceptionFormatter:
-    _default_theme = {
-        "introduction": "\x1b[33m\x1b[1m{}\x1b[0m",
-        "cause": "\x1b[1m{}\x1b[0m",
-        "context": "\x1b[1m{}\x1b[0m",
-        "dirname": "\x1b[32m{}\x1b[0m",
-        "basename": "\x1b[32m\x1b[1m{}\x1b[0m",
-        "line": "\x1b[33m{}\x1b[0m",
-        "function": "\x1b[35m{}\x1b[0m",
-        "exception_type": "\x1b[31m\x1b[1m{}\x1b[0m",
-        "exception_value": "\x1b[1m{}\x1b[0m",
-        "arrows": "\x1b[36m{}\x1b[0m",
-        "value": "\x1b[36m\x1b[1m{}\x1b[0m",
-    }
+    _default_theme = frozenset(
+        {
+            "introduction": "\x1b[33m\x1b[1m{}\x1b[0m",
+            "cause": "\x1b[1m{}\x1b[0m",
+            "context": "\x1b[1m{}\x1b[0m",
+            "dirname": "\x1b[32m{}\x1b[0m",
+            "basename": "\x1b[32m\x1b[1m{}\x1b[0m",
+            "line": "\x1b[33m{}\x1b[0m",
+            "function": "\x1b[35m{}\x1b[0m",
+            "exception_type": "\x1b[31m\x1b[1m{}\x1b[0m",
+            "exception_value": "\x1b[1m{}\x1b[0m",
+            "arrows": "\x1b[36m{}\x1b[0m",
+            "value": "\x1b[36m\x1b[1m{}\x1b[0m",
+        }.items()
+    )
 
     def __init__(
         self,
@@ -147,7 +154,7 @@ class ExceptionFormatter:
     ):
         self._colorize = colorize
         self._diagnose = diagnose
-        self._theme = theme or self._default_theme
+        self._theme = theme or dict(self._default_theme)
         self._backtrace = backtrace
         self._syntax_highlighter = SyntaxHighlighter(style)
         self._max_length = max_length
@@ -455,30 +462,31 @@ class ExceptionFormatter:
         # on the indentation; the preliminary context for "SyntaxError" is always indented, while
         # the Exception itself is not. This allows us to identify the correct index for the
         # exception message.
-        for error_message_index, part in enumerate(exception_only):  # noqa: B007
-            if not part.startswith(" "):
-                break
+        no_indented_indexes = (i for i, p in enumerate(exception_only) if not p.startswith(" "))
+        error_message_index = next(no_indented_indexes, None)
 
-        error_message = exception_only[error_message_index][:-1]  # Remove last new line temporarily
+        if error_message_index is not None:
+            # Remove final new line temporarily.
+            error_message = exception_only[error_message_index][:-1]
 
-        if self._colorize:
-            if ":" in error_message:
-                exception_type, exception_value = error_message.split(":", 1)
-                exception_type = self._theme["exception_type"].format(exception_type)
-                exception_value = self._theme["exception_value"].format(exception_value)
-                error_message = exception_type + ":" + exception_value
-            else:
-                error_message = self._theme["exception_type"].format(error_message)
+            if self._colorize:
+                if ":" in error_message:
+                    exception_type, exception_value = error_message.split(":", 1)
+                    exception_type = self._theme["exception_type"].format(exception_type)
+                    exception_value = self._theme["exception_value"].format(exception_value)
+                    error_message = exception_type + ":" + exception_value
+                else:
+                    error_message = self._theme["exception_type"].format(error_message)
 
-        if self._diagnose and frames:
-            if issubclass(exc_type, AssertionError) and not str(exc_value) and final_source:
-                if self._colorize:
-                    final_source = self._syntax_highlighter.highlight(final_source)
-                error_message += ": " + final_source
+            if self._diagnose and frames:
+                if issubclass(exc_type, AssertionError) and not str(exc_value) and final_source:
+                    if self._colorize:
+                        final_source = self._syntax_highlighter.highlight(final_source)
+                    error_message += ": " + final_source
 
-            error_message = "\n" + error_message
+                error_message = "\n" + error_message
 
-        exception_only[error_message_index] = error_message + "\n"
+            exception_only[error_message_index] = error_message + "\n"
 
         if is_first:
             yield self._prefix
@@ -497,13 +505,14 @@ class ExceptionFormatter:
             else:
                 yield from self._indent(introduction + "\n", group_nesting)
 
-        frames_lines = traceback.format_list(frames) + exception_only
+        frames_lines = self._format_list(frames) + exception_only
         if self._colorize or self._backtrace or self._diagnose:
             frames_lines = self._format_locations(frames_lines, has_introduction=has_introduction)
 
         yield from self._indent("".join(frames_lines), group_nesting)
 
         if is_grouped:
+            exc = None
             for n, exc in enumerate(value.exceptions, start=1):
                 ruler = "+" + (" %s " % ("..." if n > 15 else n)).center(35, "-")
                 yield from self._indent(ruler, group_nesting, prefix="+-" if n == 1 else "  ")
@@ -524,5 +533,41 @@ class ExceptionFormatter:
             if not is_exception_group(exc) or group_nesting == 10:
                 yield from self._indent("-" * 35, group_nesting + 1, prefix="+-")
 
+    def _format_list(self, frames):
+
+        def source_message(filename, lineno, name, line):
+            message = '  File "%s", line %d, in %s\n' % (filename, lineno, name)
+            if line:
+                message += "    %s\n" % line.strip()
+            return message
+
+        def skip_message(count):
+            plural = "s" if count > 1 else ""
+            return "  [Previous line repeated %d more time%s]\n" % (count, plural)
+
+        result = []
+        count = 0
+        last_source = None
+
+        for *source, line in frames:
+            if source != last_source and count > 3:
+                result.append(skip_message(count - 3))
+
+            if source == last_source:
+                count += 1
+                if count > 3:
+                    continue
+            else:
+                count = 1
+
+            result.append(source_message(*source, line))
+            last_source = source
+
+        # Add a final skip message if the iteration of frames ended mid-repetition.
+        if count > 3:
+            result.append(skip_message(count - 3))
+
+        return result
+
     def format_exception(self, type_, value, tb, *, from_decorator=False):
         yield from self._format_exception(value, tb, is_first=True, from_decorator=from_decorator)
